from sqlitedict import SqliteDict
import cherrypy
import cherrypy_cors

import re

count_re = re.compile(r".*_count$")


class Server(object):
    """
    Boots a server that serves the result of the report for the user. This is still a proof of concept,
    but will eventually be used to serve data when it exceeds 400mb.
    """
    def __init__(self, filename):
        """
        Constructor of the server object. Should not be called directly outside the class.

        :param filename:                Name of the file to write data to.
        :return:                        The server object.
        """
        self.results = SqliteDict(filename)

    @cherrypy.expose()
    @cherrypy.tools.json_out()
    def summary(self):
        """
        Returns the stripped down data of the results that doesn't scale up when using a lot of ressources,
        used to render the summary.
        Should be the first call from the server.
        Can be found at GET /api/summary

        :return:                        The summary data of the report.
        """
        data = dict(self.results)
        services = data.get('services')
        stripped_services = {}
        for k1, v1 in services.items():
            service = {}
            for k2, v2 in v1.items():
                if k2 == 'findings' or k2 == 'filters' or count_re.match(k2):
                    service[k2] = v2
            stripped_services[k1] = service
        data['services'] = stripped_services
        return {'data': data}

    @cherrypy.expose()
    @cherrypy.tools.json_out()
    def data(self, key=None):
        """
        Return the data at the requested key. Doesn't returns nested dictionnaries and lists.
        If one of the value is a dictionnary, it will return {'type': 'dict', 'keys': <Array of all the keys>}
        If one of the value is a list, it will return {'type': 'list', 'count': <number of elements in the list>}

        Can be found at GET /api/data?key=<KEY>
        :param key:                     Key of the requested information, separated by the character '¤'.
        :return:                        The data at the requested location stripped of its nested data.
        """
        result = self.get_item(self.results, key)
        # Returns only indexes or length if it's a complex type
        if isinstance(result, dict) or isinstance(result, SqliteDict):
            result = {'type': 'dict', 'keys': list(result.keys())}
        elif isinstance(result, list):
            result = {'type': 'list', 'length': len(result)}
        return {'data': result}

    @cherrypy.expose()
    @cherrypy.tools.json_out()
    def full(self, key=None):
        """
        Return the data at the requested key. Returns all the nested data.
        Be sure not to use it on a key that may contains a lot of data, as the request won't be answered
        if it's too large(generally 3mb).

        Can be found at GET /api/full?key=<KEY>
        :param key:                     Key of the requested information, separated by the character '¤'.
        :return:                        The data at the requested location.
        """
        result = self.get_item(self.results, key)
        if isinstance(result, str) or isinstance(result, int):
            return {'data': result}
        return {'data': dict(result)}

    @cherrypy.expose()
    @cherrypy.tools.json_out()
    def page(self, key=None, page=None, pagesize=None):
        """
        Return a page of the data at the requested key. Doesn't returns nested dictionnaries and lists.
        For example, if you set pagesize=10 and page=2, it should return element 10-19
        If one of the value is a dictionnary, it will return {'type': 'dict', 'keys': <Array of all the keys>}
        If one of the value is a list, it will return {'type': 'list', 'count': <number of elements in the list>}

        Can be found at GET /api/page?key=<KEY>&page=<PAGE>&pagesize=<PAGESIZE>
        :param key:                     Key of the requested information, separated by the character '¤'.
        :param page:                    The number of the page you request.
        :param pagesize:                The size of the page you request.
        :return:                        A subset of the data at the requested location.
        """
        result = self.get_item(self.results, key)

        page = int(page)
        pagesize = int(pagesize)

        start = page * pagesize
        end = min((page + 1) * pagesize, len(result))

        if isinstance(result, dict) or isinstance(result, SqliteDict):
            page = {k: result.get(k) for k in sorted(list(result))[start:end]}
        if isinstance(result, list):
            page = result[start:end]

        return {'data': self.strip_nested_data(page)}

    @staticmethod
    def init(database_filename, host, port):
        """
        Configure and starts the server.

        :param database_filename:       Location of the database file.
        :param host:                    Address on which to listen.
        :param port:                    Port on which to listen.
        """
        cherrypy_cors.install()
        config = {
            '/': {
                'cors.expose.on': True,
                'tools.sessions.on': True,
                'tools.response_headers.on': True,
                'tools.response_headers.headers': [('Content-Type', 'text/plain')],
            },
        }
        cherrypy.config.update({
                'server.socket_host': host,
                'server.socket_port': port,
        })
        cherrypy.quickstart(Server(database_filename), "/api", config=config)

    @staticmethod
    def get_item(data, key):
        """
        Get a specific informations from its key.

        :param data:                    The dictionnary in which the information is stored.
        :param host:                    The key where the information is located.
        :return:                        The nested data at the requested location.
        """
        if not key:
            return data

        keyparts = key.split('¤')
        for k in keyparts:
            if isinstance(data, dict) or isinstance(data, SqliteDict):
                data = data.get(k)
            elif isinstance(data, list):
                data = data[int(k)]
        return data

    @staticmethod
    def strip_nested_data(data):
        """
        Strip nested lists and dictionnaries from the provided object to reduce its size.

        :param data:                    The object to strip.
        :return:                        The input data stripped of its nested lists and dictionnaries.
        """
        if not isinstance(data, dict):
            return data

        result = {}
        for k, v in data.items():
            if isinstance(v, dict):
                result[k] = {'type': 'dict', 'keys': list(v.keys())}
            elif isinstance(v, list):
                result[k] = {'type': 'list', 'length': len(v)}
        return result




from __future__ import print_function

formatted_service_name = {
    # AWS
    'cloudformation': 'CloudFormation',
    'cloudtrail': 'CloudTrail',
    'cloudwatch': 'CloudWatch',
    'credentials': 'Credentials',
    'config': 'Config',
    'directconnect': 'Direct Connect',
    'dynamodb': 'DynamoDB',
    'elbv2': 'ELBv2',
    'elasticache': 'ElastiCache',
    'lambda': 'Lambda',
    'awslambda': 'Lambda',
    'redshift': 'RedShift',
    'route53': 'Route53',
    # Azure
    'storageaccounts': 'Storage Accounts',
    'sqldatabase': 'SQL Database',
    'securitycenter': 'Security Center',
    'keyvault': 'Key Vault',
    'appgateway': 'Application Gateway',
    'rediscache': 'Redis Cache',
    'network': 'Network',
    'appservice': 'App Service',
    'loadbalancer': 'Load Balancer',
    'virtualmachines': 'Virtual Machines',
    'graphrbac': 'Graph RBAC',
    # GCP
    'cloudstorage': 'Cloud Storage',
    'cloudsql': 'Cloud SQL',
    'stackdriverlogging': 'Stackdriver Logging',
    'stackdrivermonitoring': 'Stackdriver Monitoring',
    'computeengine': 'Compute Engine',
    'kubernetesengine': 'Kubernetes Engine',
    'cloudresourcemanager': 'Cloud Resource Manager',
    # Aliyun
    'actiontrail': 'ActionTrail',
    # OCI
    'identity': 'Identity',
    'objectstorage': 'Object Storage',
}


def manage_dictionary(dictionary, key, init, callback=None):
    """
    :param dictionary:
    :param key:
    :param init:
    :param callback:
    :return:
    """

    if str(key) in dictionary:
        return dictionary

    dictionary[str(key)] = init
    manage_dictionary(dictionary, key, init)
    if callback:
        callback(dictionary[key])
    return dictionary


def format_service_name(service):
    """

    :param service:
    :return:
    """
    return formatted_service_name[service] if service in formatted_service_name else service.upper()



from ScoutSuite.providers.aws.resources.base import AWSResources


class Groups(AWSResources):
    async def fetch_all(self):
        raw_groups = await self.facade.iam.get_groups()
        for raw_group in raw_groups:
            name, resource = self._parse_group(raw_group)
            self[name] = resource

    def _parse_group(self, raw_group):
        if raw_group['GroupName'] in self:
            return

        raw_group['id'] = raw_group.pop('GroupId')
        raw_group['name'] = raw_group.pop('GroupName')
        raw_group['arn'] = raw_group.pop('Arn')
        raw_group['users'] = raw_group.pop('Users')
        return raw_group['id'], raw_group



from ScoutSuite.providers.aliyun.facade.base import AliyunFacade
from ScoutSuite.providers.base.services import BaseServicesConfig
from ScoutSuite.providers.aliyun.resources.ram.base import RAM
from ScoutSuite.providers.aliyun.resources.actiontrail.base import ActionTrail
from ScoutSuite.providers.aliyun.resources.vpc.base import VPC
from ScoutSuite.providers.aliyun.resources.ecs.base import ECS
from ScoutSuite.providers.aliyun.resources.rds.base import RDS
from ScoutSuite.providers.aliyun.resources.kms.base import KMS



class AliyunServicesConfig(BaseServicesConfig):
    def __init__(self, credentials, **kwargs):
        super(AliyunServicesConfig, self).__init__(credentials)

        facade = AliyunFacade(credentials)

        self.actiontrail = ActionTrail(facade)
        self.ram = RAM(facade)
        self.ecs = ECS(facade)
        self.rds = RDS(facade)
        self.vpc = VPC(facade)
        self.kms = KMS(facade)

    def _is_provider(self, provider_name):
        return provider_name == 'aliyun'



from ScoutSuite.providers.aliyun.authentication_strategy import AliyunCredentials
from ScoutSuite.providers.aliyun.facade.utils import get_response
from ScoutSuite.providers.aliyun.utils import get_client

from ScoutSuite.core.console import print_exception
from aliyunsdkvpc.request.v20160428 import DescribeVpcsRequest


class VPCFacade:
    def __init__(self, credentials: AliyunCredentials):
        self._credentials = credentials

    async def get_vpcs(self, region):
        """
        Get all VPCs

        :return: a list of all VPCs
        """
        client = get_client(credentials=self._credentials, region=region)
        response = await get_response(client=client,
                                      request=DescribeVpcsRequest.DescribeVpcsRequest())
        if response:
            return response['Vpcs']['Vpc']
        else:
            return []



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.utils import run_concurrently


class ConfigFacade(AWSBaseFacade):

    async def get_rules(self, region):
        try:
            return await AWSFacadeUtils.get_all_pages('config', region, self.session, 'describe_config_rules', 'ConfigRules')
        except Exception as e:
            print_exception('Failed to get Config ruless: {}'.format(e))
            return []

    async def get_recorders(self, region: str):
        client = AWSFacadeUtils.get_client('config', self.session, region)

        try:
            recorders = (await run_concurrently(client.describe_configuration_recorders))['ConfigurationRecorders']
        except Exception as e:
            print_exception('Failed to get Config recorders: {}'.format(e))
            recorders = []

        try:
            recorder_statuses_list = \
                (await run_concurrently(client.describe_configuration_recorder_status))['ConfigurationRecordersStatus']
        except Exception as e:
            print_exception('Failed to get Config recorder statuses: {}'.format(e))
        else:
            # To accelerate the mapping of the statuses, we preprocess the data by creating a
            # <recorder_name: recorder_status> map. This prevents having to iterate over the list of statuses for each
            # recorder.
            recorder_statuses_map = {recorder['name']: recorder for recorder in recorder_statuses_list}
            for recorder in recorders:
                recorder['ConfigurationRecordersStatus'] = recorder_statuses_map[recorder['name']]

        return recorders



from ScoutSuite.providers.aliyun.resources.base import AliyunResources
from ScoutSuite.providers.aliyun.facade.base import AliyunFacade


class Keys(AliyunResources):
    def __init__(self, facade: AliyunFacade, region: str):
        super(Keys, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        for raw_key in await self.facade.kms.get_keys(region=self.region):
            id, key = await self._parse_key(raw_key)
            self[id] = key

    async def _parse_key(self, raw_key):
        key_dict = {}
        key_dict['id'] = raw_key.get('KeyId')
        key_dict['name'] = raw_key.get('KeyId')
        key_dict['arn'] = raw_key.get('KeyArn')

        # get additional details for the key
        raw_key_details = await self.facade.kms.get_key_details(key_dict['id'], region=self.region)

        key_dict['creation_date'] = raw_key_details.get('CreationDate')
        key_dict['delete_date'] = raw_key_details.get('DeleteDate')
        key_dict['origin'] = raw_key_details.get('Origin')
        key_dict['description'] = raw_key_details.get('Description')
        key_dict['creator'] = raw_key_details.get('Creator')
        key_dict['usage'] = raw_key_details.get('KeyUsage')
        key_dict['material_expire_time'] = raw_key_details.get('MaterialExpireTime')
        key_dict['state'] = raw_key_details.get('KeyState')

        if key_dict['delete_date'] == '':
            key_dict['delete_date'] = None
        if key_dict['material_expire_time'] == '':
            key_dict['material_expire_time'] = None

        return key_dict['id'], key_dict




import json

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import get_and_set_concurrently
from ScoutSuite.providers.utils import run_concurrently


class CloudFormation(AWSBaseFacade):

    async def get_stacks(self, region: str):
        try:
            stacks = await AWSFacadeUtils.get_all_pages(
                'cloudformation', region, self.session, 'list_stacks', 'StackSummaries')
        except Exception as e:
            print_exception('Failed to get CloudFormation stack: {}'.format(e))
            stacks = []
        else:
            stacks = [stack for stack in stacks if not CloudFormation._is_stack_deleted(stack)]
            await get_and_set_concurrently(
                [self._get_and_set_description, self._get_and_set_template, self._get_and_set_policy],
                stacks, region=region)
        finally:
            return stacks

    async def _get_and_set_description(self, stack: {}, region: str):
        client = AWSFacadeUtils.get_client('cloudformation', self.session, region)
        try:
            stack_description = await run_concurrently(
                lambda: client.describe_stacks(StackName=stack['StackName'])['Stacks'][0])
        except Exception as e:
            print_exception('Failed to describe CloudFormation stack: {}'.format(e))
        else:
            stack.update(stack_description)

    async def _get_and_set_template(self, stack: {}, region: str):
        client = AWSFacadeUtils.get_client('cloudformation', self.session, region)
        try:
            stack['template'] = await run_concurrently(
                lambda: client.get_template(StackName=stack['StackName'])['TemplateBody'])
        except Exception as e:
            print_exception('Failed to get CloudFormation template: {}'.format(e))
            stack['template'] = None

    async def _get_and_set_policy(self, stack: {}, region: str):
        client = AWSFacadeUtils.get_client('cloudformation', self.session, region)
        try:
            stack_policy = await run_concurrently(
                lambda: client.get_stack_policy(StackName=stack['StackName']))
        except Exception as e:
            print_exception('Failed to get CloudFormation stack policy: {}'.format(e))
        else:
            if 'StackPolicyBody' in stack_policy:
                stack['policy'] = json.loads(stack_policy['StackPolicyBody'])

    @staticmethod
    def _is_stack_deleted(stack):
        return stack.get('StackStatus', None) == 'DELETE_COMPLETE'



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from .buckets import Buckets


class S3(AWSCompositeResources):
    _children = [
        (Buckets, 'buckets')
    ]

    def __init__(self, facade: AWSFacade):
        super(S3, self).__init__(facade)
        self.service = 's3'

    async def fetch_all(self, partition_name='aws', **kwargs):
        await self._fetch_children(self)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources

from .clusters import EMRClusters


class EMRVpcs(AWSCompositeResources):
    _children = [
        (EMRClusters, 'clusters')
    ]

    def __init__(self, facade: AWSFacade, region: str):
        self.region = region

        super(EMRVpcs, self).__init__(facade)

    async def fetch_all(self):
        # EMR won't disclose its VPC, so we put everything in a VPC named "EMR-UNKNOWN-VPC", and we
        # infer the VPC afterwards during the preprocessing.
        tmp_vpc = 'EMR-UNKNOWN-VPC'
        self[tmp_vpc] = {}
        await self._fetch_children(self[tmp_vpc], {'region': self.region})



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class Snapshots(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super(Snapshots, self).__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_snapshots = await self.facade.gce.get_snapshots(self.project_id)
        for raw_snapshot in raw_snapshots:
            snapshot_id, snapshot = self._parse_snapshot(raw_snapshot)
            self[snapshot_id] = snapshot

    def _parse_snapshot(self, raw_snapshot):
        snapshot_dict = {}
        snapshot_dict['id'] = raw_snapshot['id']
        snapshot_dict['name'] = raw_snapshot['name']
        snapshot_dict['description'] = self._get_description(raw_snapshot)
        snapshot_dict['creation_timestamp'] = raw_snapshot['creationTimestamp']
        snapshot_dict['status'] = raw_snapshot['status']
        snapshot_dict['source_disk_id'] = raw_snapshot['sourceDiskId']
        snapshot_dict['source_disk_url'] = raw_snapshot['sourceDisk']
        return snapshot_dict['id'], snapshot_dict

    def _get_description(self, raw_snapshot):
        description = raw_snapshot.get('description')
        return description if description else 'N/A'



import os

from ScoutSuite.core.console import print_exception, print_info
from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.providers.gcp.services import GCPServicesConfig


class GCPProvider(BaseProvider):
    """
    Implements provider for GCP
    """

    def __init__(self,
                 project_id=None, folder_id=None, organization_id=None, all_projects=None,
                 report_dir=None, timestamp=None, services=None, skipped_services=None, result_format='json', **kwargs):
        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(
            os.path.abspath(__file__))[0]

        self.provider_code = 'gcp'
        self.provider_name = 'Google Cloud Platform'
        self.environment = 'default'

        self.all_projects = all_projects
        self.project_id = project_id
        self.folder_id = folder_id
        self.organization_id = organization_id

        self.credentials = kwargs['credentials']
        self._set_account_id()

        self.services = GCPServicesConfig(self.credentials, self.credentials.default_project_id,
                                          self.project_id, self.folder_id, self.organization_id, self.all_projects)

        self.result_format = result_format

        super(GCPProvider, self).__init__(report_dir, timestamp,
                                          services, skipped_services, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        if self.account_id:
            return 'gcp-{}'.format(self.account_id)
        else:
            return 'gcp'

    def _set_account_id(self):
        # All accessible projects
        if self.all_projects:
            # Service Account
            if self.credentials.is_service_account and hasattr(self.credentials, 'service_account_email'):
                self.account_id = self.credentials.service_account_email
            else:
                # TODO use username email (can't find it...)
                self.account_id = 'user-account'
        # Project passed through the CLI
        elif self.project_id:
            self.account_id = self.project_id
        # Folder passed through the CLI
        elif self.folder_id:
            self.account_id = self.folder_id
        # Organization passed through the CLI
        elif self.organization_id:
            self.account_id = self.organization_id
        # Project inferred from default configuration
        elif self.credentials.default_project_id:
            self.account_id = self.credentials.default_project_id
        else:
            self.account_id = 'unknown-project-id'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        """
        Tweak the GCP config to match cross-resources and clean any fetching artifacts

        :param ip_ranges:
        :param ip_ranges_name_key:
        :return: None
        """
        ip_ranges = [] if ip_ranges is None else ip_ranges

        self._match_instances_and_snapshots()
        self._match_networks_and_instances()

        super(GCPProvider, self).preprocessing()

    def _match_instances_and_snapshots(self):
        """
        Compare Compute Engine instances and snapshots to identify instance disks that do not have a snapshot.

        :return:
        """

        if 'computeengine' in self.service_list:
            for project in self.services['computeengine']['projects'].values():
                for zone in project['zones'].values():
                    for instance in zone['instances'].values():
                        for instance_disk in instance['disks'].values():
                            instance_disk['snapshots'] = []
                            for disk in project['snapshots'].values():
                                if disk['status'] == 'READY' and disk['source_disk_url'] == instance_disk['source_url']:
                                    instance_disk['snapshots'].append(disk)

                            instance_disk['latest_snapshot'] = max(instance_disk['snapshots'],
                                                                   key=lambda x: x['creation_timestamp']) \
                                if instance_disk['snapshots'] else None

    def _match_networks_and_instances(self):
        """
        For each network, math instances in that network

        :return:
        """

        if 'computeengine' in self.service_list:
            for project in self.services['computeengine']['projects'].values():
                for network in project['networks'].values():
                    network['instances'] = []
                    for zone in project['zones'].values():
                        # Skip the counts contained in the zones dictionary
                        if zone is int:
                            continue
                        for instance in zone['instances'].values():
                            for network_interface in instance['network_interfaces']:
                                if network_interface['network'] == network['network_url']:
                                    network['instances'].append(instance['id'])



from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Watchers(AzureResources):
    async def fetch_all(self):
        for raw_watcher in await self.facade.network.get_network_watchers():
            id, network_watcher = self._parse_network_watcher(raw_watcher)
            self[id] = network_watcher

    def _parse_network_watcher(self, raw_watcher):
        watcher_dict = {}
        watcher_dict['id'] = get_non_provider_id(raw_watcher.id)
        watcher_dict['name'] = raw_watcher.name
        watcher_dict['type'] = raw_watcher.type
        watcher_dict['location'] = raw_watcher.location
        watcher_dict['tags'] = raw_watcher.tags
        watcher_dict['etag'] = raw_watcher.etag
        watcher_dict['additional_properties'] = raw_watcher.additional_properties
        watcher_dict['provisioning_state'] = raw_watcher.provisioning_state
        return watcher_dict['id'], watcher_dict




import os

from ScoutSuite.providers.azure.services import AzureServicesConfig
from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.providers.azure.services import AzureServicesConfig


class AzureProvider(BaseProvider):
    """
    Implements provider for Azure
    """

    def __init__(self, project_id=None, organization_id=None,
                 report_dir=None, timestamp=None, services=None, skipped_services=None,
                 result_format='json', **kwargs):
        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(os.path.abspath(__file__))[0]

        self.provider_code = 'azure'
        self.provider_name = 'Microsoft Azure'
        self.environment = 'default'

        self.services_config = AzureServicesConfig

        self.credentials = kwargs['credentials']
        self.account_id = self.credentials.subscription_id
        self.result_format = result_format

        super(AzureProvider, self).__init__(report_dir, timestamp,
                                            services, skipped_services, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        if self.credentials.subscription_id:
            return 'azure-{}'.format(self.credentials.subscription_id)
        elif self.account_id:
            return 'azure-{}'.format(self.account_id)
        else:
            return 'azure'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        """
        Tweak the Azure config to match cross-service resources and clean any fetching artifacts

        :param ip_ranges:
        :param ip_ranges_name_key:
        :return: None
        """
        ip_ranges = [] if ip_ranges is None else ip_ranges

        super(AzureProvider, self).preprocessing()



from aliyunsdkcore.client import AcsClient
from aliyunsdksts.request.v20150401 import GetCallerIdentityRequest
import json

from ScoutSuite.core.console import print_exception


def get_client(credentials, region=None):
    try:
        client = AcsClient(credential=credentials.credentials, region_id=region if region else 'cn-hangzhou')
        return client

    except Exception as e:
        print_exception(e)
        return None



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class Subnetworks(Resources):
    def __init__(self, facade: GCPFacade, project_id: str, region: str):
        super(Subnetworks, self).__init__(facade)
        self.project_id = project_id
        self.region = region

    async def fetch_all(self):
        raw_subnetworks = await self.facade.gce.get_subnetworks(self.project_id, self.region)
        for raw_subnetwork in raw_subnetworks:
            subnetwork_id, subnetwork = self._parse_subnetwork(raw_subnetwork)
            self[subnetwork_id] = subnetwork

    def _parse_subnetwork(self, raw_subnetwork):
        subnetwork_dict = {}
        subnetwork_dict['id'] = raw_subnetwork['id']
        subnetwork_dict['project_id'] = raw_subnetwork['selfLink'].split('/')[-5]
        subnetwork_dict['region'] = raw_subnetwork['region'].split('/')[-1]
        subnetwork_dict['name'] = "%s-%s" % (raw_subnetwork['name'], subnetwork_dict['region'])
        subnetwork_dict['subnetwork'] = raw_subnetwork['network'].split('/')[-1]
        subnetwork_dict['gateway_address'] = raw_subnetwork['gatewayAddress']
        subnetwork_dict['ip_range'] = raw_subnetwork['ipCidrRange']
        subnetwork_dict['creation_timestamp'] = raw_subnetwork['creationTimestamp']
        subnetwork_dict['private_ip_google_access'] = raw_subnetwork['privateIpGoogleAccess']
        return subnetwork_dict['id'], subnetwork_dict



from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.base.resources.base import Resources


class Bindings(Resources):
    def __init__(self, facade: GCPFacade, project_id: str, service_account_email: str):
        super(Bindings, self).__init__(facade)
        self.project_id = project_id
        self.service_account_email = service_account_email 

    async def fetch_all(self):
        raw_bindings = await self.facade.iam.get_bindings(self.project_id, self.service_account_email)
        for raw_binding in raw_bindings:
            binding_id, binding = self._parse_binding(raw_binding)
            self[binding_id] = binding

    def _parse_binding(self, raw_binding):
        return len(self), raw_binding



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class PeeringConnections(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        self.facade = facade
        self.region = region

    async def fetch_all(self):
        raw_peering_connections = await self.facade.ec2.get_peering_connections(self.region)
        for raw_peering_connection in raw_peering_connections:
            id, peering_connection = self._parse_peering_connections(raw_peering_connection)
            self[id] = peering_connection

    def _parse_peering_connections(self, raw_peering_connection):
        peering_connection_id = raw_peering_connection['VpcPeeringConnectionId']
        return peering_connection_id, raw_peering_connection



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.iam.service_accounts import ServiceAccounts


class IAM(Projects):
    _children = [ 
        (ServiceAccounts, 'service_accounts') 
    ]



from ScoutSuite.core.console import print_debug

from ScoutSuite.output.result_encoder import JavaScriptEncoder


class RuleExceptions(object):
    """
    Exceptions handling
    """

    def __init__(self, file_path=None):
        self.jsrw = JavaScriptEncoder()
        self.exceptions = self.jsrw.load_from_file(file_type='EXCEPTIONS',
                                                   file_path=file_path,
                                                   first_line=True)

    def process(self, cloud_provider):
        for service in self.exceptions:
            for rule in self.exceptions[service]:
                filtered_items = []
                if rule not in cloud_provider.services[service]['findings']:
                    print_debug('Warning:: key error should not be happening')
                    continue
                for item in cloud_provider.services[service]['findings'][rule]['items']:
                    if item not in self.exceptions[service][rule]:
                        filtered_items.append(item)
                cloud_provider.services[service]['findings'][rule]['items'] = filtered_items
                cloud_provider.services[service]['findings'][rule]['flagged_items'] = \
                    len(cloud_provider.services[service]['findings'][rule]['items'])



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .identities import Identities


class SES(Regions):
    _children = [
        (Identities, 'identities')
    ]

    def __init__(self, facade: AWSFacade):
        super(SES, self).__init__('ses', facade)



from oci.key_management import KmsManagementClient, KmsVaultClient
from oci.pagination import list_call_get_all_results

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.oci.authentication_strategy import OracleCredentials
from ScoutSuite.providers.utils import run_concurrently


class KMSFacade:
    def __init__(self, credentials: OracleCredentials):
        self._credentials = credentials
        # FIXME does this require regional support?
        self._vault_client = KmsVaultClient(self._credentials.config)

    async def get_vaults(self):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._vault_client.list_vaults, self._credentials.compartment_id))
            return response.data
        except Exception as e:
            print_exception('Failed to get KMS vaults: {}'.format(e))
            return []

    async def get_keys(self, keyvault):
        try:
            key_client = KmsManagementClient(self._credentials.config, keyvault['management_endpoint'])
            response = await run_concurrently(
                lambda: list_call_get_all_results(key_client.list_keys, self._credentials.compartment_id))
            return response.data
        except Exception as e:
            print_exception('Failed to get KMS vaults: {}'.format(e))
            return []



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.providers.utils import run_concurrently

class IAMFacade(GCPBaseFacade):
    def __init__(self):
        super(IAMFacade, self).__init__('iam', 'v1')

    async def get_bindings(self, project_id: str, service_account_email: str):
        try:
            resource = 'projects/{}/serviceAccounts/{}'.format(project_id, service_account_email)
            iam_client = self._get_client()
            response = await run_concurrently(
                    lambda: iam_client.projects().serviceAccounts().getIamPolicy(resource=resource).execute()
            )
            return response.get('bindings', [])
        except Exception as e:
            print_exception('Failed to retrieve service account IAM policy bindings: {}'.format(e))
            return []

    async def get_keys(self, project_id: str, service_account_email: str, key_types: list=[]):
        try:
            name = 'projects/{}/serviceAccounts/{}'.format(project_id, service_account_email)
            iam_client = self._get_client()
            response = await run_concurrently(
                    lambda: iam_client.projects().serviceAccounts().keys().list(name=name,
                                                                                keyTypes=key_types).execute()
            )
            return response.get('keys', [])
        except Exception as e:
            print_exception('Failed to retrieve service account keys: {}'.format(e))
            return []

    async def get_key(self, key_name: str):
        try:
            iam_client = self._get_client()
            response = await run_concurrently(
                lambda: iam_client.projects().serviceAccounts().keys().get(name=key_name,
                                                                           fields='').execute()
            )
            return response
        except Exception as e:
            print_exception('Failed to retrieve service account keys: {}'.format(e))
            return []

    async def get_service_accounts(self, project_id: str):
        try:
            name = 'projects/{}'.format(project_id)     
            iam_client = self._get_client()
            request = iam_client.projects().serviceAccounts().list(name=name)
            service_accounts_group = iam_client.projects().serviceAccounts()
            return await GCPFacadeUtils.get_all('accounts', request, service_accounts_group)
        except Exception as e:
            print_exception('Failed to retrieve service accounts: {}'.format(e))
            return []



from ScoutSuite.providers.gcp.resources.gce.subnetworks import Subnetworks
from ScoutSuite.providers.gcp.resources.regions import Regions


class GCERegions(Regions):
    _children = [
        (Subnetworks, 'subnetworks')
    ]



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .functions import Functions


class Lambdas(Regions):
    _children = [
        (Functions, 'functions')
    ]

    def __init__(self, facade: AWSFacade):
        super(Lambdas, self).__init__('lambda', facade)



from ScoutSuite.providers.gcp.resources.gce.disks import Disks


class InstanceDisks(Disks):
    def __init__(self, facade, instance):
        super(InstanceDisks, self).__init__(facade)
        self.instance = instance

    def fetch_all(self):
        raw_disks = self.instance.get('disks', {})
        for raw_disk in raw_disks:
            disk_id, disk = self._parse_disk(raw_disk)
            self[disk_id] = disk
        # We need self.instance to get the disks, but we do 
        # not want to have it in the generated JSON.
        del self.instance



"""This module provides implementations for Resources and CompositeResources for Aliyun."""

import abc

from ScoutSuite.providers.base.resources.base import Resources, CompositeResources


class AliyunResources(Resources, metaclass=abc.ABCMeta):
    """This is the base class for Aliyun resources."""

    pass


class AliyunCompositeResources(AliyunResources, CompositeResources, metaclass=abc.ABCMeta):
    """This class represents a collection of composite Resources (resources that include nested resources referred as
    their children). Classes extending AliyunCompositeResources have to define a '_children' attribute which consists of
    a list of tuples describing the children. The tuples are expected to respect the following format:
    (<child_class>, <child_name>). 'child_name' is used to indicate the name under which the child resources will be
    stored in the parent object.
    """

    pass



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Functions(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(Functions, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_functions = await self.facade.awslambda.get_functions(self.region)
        for raw_function in raw_functions:
            name, resource = self._parse_function(raw_function)
            self[name] = resource

    def _parse_function(self, raw_function):
        raw_function['name'] = raw_function.pop('FunctionName')
        return raw_function['name'], raw_function



__author__ = 'NCC Group'
__version__ = '5.3.1'

ERRORS_LIST = []

DEFAULT_REPORT_DIRECTORY = 'scoutsuite-report'
DEFAULT_REPORT_RESULTS_DIRECTORY = 'scoutsuite-results'
DEFAULT_INCLUDES_DIRECTORY = 'inc-scoutsuite'



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class SubnetGroups(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super(SubnetGroups, self).__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_subnet_groups = await self.facade.elasticache.get_subnet_groups(self.region, self.vpc)
        for raw_subnet_group in raw_subnet_groups:
            name, resource = self._parse_subnet_group(raw_subnet_group)
            self[name] = resource

    def _parse_subnet_group(self, raw_subnet_group):
        raw_subnet_group['name'] = raw_subnet_group.pop('CacheSubnetGroupName')
        return raw_subnet_group['name'], raw_subnet_group



from oci.object_storage import ObjectStorageClient
from ScoutSuite.providers.oci.authentication_strategy import OracleCredentials
from oci.pagination import list_call_get_all_results

from ScoutSuite.providers.utils import run_concurrently


class ObjectStorageFacade:
    def __init__(self, credentials: OracleCredentials):
        self._credentials = credentials
        self._client = ObjectStorageClient(self._credentials.config)

    async def get_namespace(self):
        response = await run_concurrently(
            lambda: list_call_get_all_results(self._client.get_namespace))
        # for some reason it returns a list of chars instead of a string
        return ''.join(response.data)

    async def get_bucket_details(self, namespace, bucket_name):
        response = await run_concurrently(
            lambda: self._client.get_bucket(namespace, bucket_name))
        return response.data

    async def get_buckets(self, namespace):
        response = await run_concurrently(
            lambda: list_call_get_all_results(self._client.list_buckets, namespace, self._credentials.compartment_id))
        return response.data

    async def get_bucket_objects(self, namespace, bucket_name):
        response = await run_concurrently(
            lambda: list_call_get_all_results(self._client.list_objects, namespace, bucket_name))
        return response.data



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Snapshots(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super(Snapshots, self).__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_snapshots = await self.facade.rds.get_snapshots(self.region, self.vpc)
        for raw_snapshot in raw_snapshots:
            name, resource = self._parse_snapshot(raw_snapshot)
            self[name] = resource

    def _parse_snapshot(self, raw_snapshot):
        snapshot_id = raw_snapshot.pop('DBSnapshotIdentifier')
        snapshot = {}
        snapshot['arn'] = raw_snapshot.pop('DBSnapshotArn')
        snapshot['id'] = snapshot_id,
        snapshot['name'] = snapshot_id,
        snapshot['vpc_id'] = raw_snapshot['VpcId']
        snapshot['attributes'] = raw_snapshot['Attributes']

        attributes = [
            'DBInstanceIdentifier',
            'SnapshotCreateTime',
            'Encrypted',
            'OptionGroupName'
        ]
        for attribute in attributes:
            snapshot[attribute] = raw_snapshot[attribute] if attribute in raw_snapshot else None

        return snapshot_id, snapshot



import os

from ScoutSuite.providers.aliyun.services import AliyunServicesConfig
from ScoutSuite.providers.base.provider import BaseProvider


class AliyunProvider(BaseProvider):
    """
    Implements provider for Azure
    """

    def __init__(self,
                 report_dir=None, timestamp=None, services=None, skipped_services=None, **kwargs):

        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(os.path.abspath(__file__))[0]

        self.provider_code = 'aliyun'
        self.provider_name = 'Alibaba Cloud'
        self.environment = 'default'

        self.services_config = AliyunServicesConfig

        self.credentials = kwargs['credentials']
        self.account_id = self.credentials.caller_details['AccountId']

        super(AliyunProvider, self).__init__(report_dir, timestamp, services, skipped_services)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        if self.account_id:
            return 'aliyun-{}'.format(self.account_id)
        else:
            return 'aliyun'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):

        super(AliyunProvider, self).preprocessing()




import re

ec2_classic = 'EC2-Classic'


def get_caller_identity(session):
    sts_client = session.client('sts')
    identity = sts_client.get_caller_identity()
    return identity


def get_aws_account_id(session):
    caller_identity = get_caller_identity(session)
    account_id = caller_identity['Arn'].split(':')[4]
    return account_id


def get_partition_name(session):
    caller_identity = get_caller_identity(session)
    partition_name = caller_identity['Arn'].split(':')[1]
    return partition_name


def is_throttled(e):
    """
    Determines whether the exception is due to API throttling.

    :param e:                           Exception raised
    :return:                            True if it's a throttling exception else False
    """
    return (hasattr(e, 'response')
            and e.response
            and 'Error' in e.response
            and e.response['Error']['Code'] in ['Throttling',
                                                'RequestLimitExceeded',
                                                'ThrottlingException'])


def get_keys(src, dst, keys):
    """
    Copies the value of keys from source object to dest object

    :param src:                         Source object
    :param dst:                         Destination object
    :param keys:                        Keys
    :return:
    """
    for key in keys:
        dst[key] = src[key] if key in src else None


def get_name(src, dst, default_attribute):
    """

    :param src:                         Source object
    :param dst:                         Destination object
    :param default_attribute:           Default attribute

    :return:
    """
    name_found = False
    if 'Tags' in src:
        for tag in src['Tags']:
            if tag['Key'] == 'Name' and tag['Value'] != '':
                dst['name'] = tag['Value']
                name_found = True
    if not name_found:
        dst['name'] = src[default_attribute]
    return dst['name']


def no_camel(name):
    """
    Converts CamelCase to camel_case

    :param name:                        Name string to convert
    :return:
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



import boto3

from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy, AuthenticationException
from ScoutSuite.providers.aws.utils import get_caller_identity


class AWSCredentials:

    def __init__(self, session):
        self.session = session


class AWSAuthenticationStrategy(AuthenticationStrategy):
    """
    Implements authentication for the AWS provider
    """

    def authenticate(self,
                     profile=None,
                     aws_access_key_id=None, aws_secret_access_key=None, aws_session_token=None,
                     **kwargs):

        try:

            if profile:
                session = boto3.Session(profile_name=profile)
            elif aws_access_key_id and aws_secret_access_key:
                if aws_session_token:
                    session = boto3.Session(
                        aws_access_key_id=aws_access_key_id,
                        aws_secret_access_key=aws_secret_access_key,
                        aws_session_token=aws_session_token,
                    )
                else:
                    session = boto3.Session(
                        aws_access_key_id=aws_access_key_id,
                        aws_secret_access_key=aws_secret_access_key,
                    )
            else:
                raise AuthenticationException('Insufficient credentials provided')

            # Test querying for current user
            identity = get_caller_identity(session)

            return AWSCredentials(session=session)

        except Exception as e:
            raise AuthenticationException(e)



from ScoutSuite.providers.aliyun.resources.base import AliyunResources
from ScoutSuite.providers.aliyun.facade.base import AliyunFacade


class Instances(AliyunResources):
    def __init__(self, facade: AliyunFacade, region: str):
        super(Instances, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        for raw_instance in await self.facade.rds.get_instances(region=self.region):
            id, instance = await self._parse_instance(raw_instance)
            self[id] = instance

    async def _parse_instance(self, raw_instance):
        instance_dict = {}

        instance_dict['id'] = raw_instance.get('DBInstanceId')
        instance_dict['name'] = raw_instance.get('DBInstanceDescription', raw_instance.get('DBInstanceId'))
        instance_dict['create_time'] = raw_instance.get('CreateTime')
        instance_dict['expire_time'] = raw_instance.get('ExpireTime')
        instance_dict['ins_id'] = raw_instance.get('InsId')
        instance_dict['lock_mode'] = raw_instance.get('LockMode')
        instance_dict['db_instance_net_type'] = raw_instance.get('DBInstanceNetType')
        instance_dict['read_only_db_instance_ids'] = raw_instance.get('ReadOnlyDBInstanceIds')
        instance_dict['lock_reason'] = raw_instance.get('LockReason')
        instance_dict['engine'] = raw_instance.get('Engine')
        instance_dict['vpc_id'] = raw_instance.get('VpcId')
        instance_dict['mutri_o_rsignle'] = raw_instance.get('MutriORsignle')
        instance_dict['connection_mode'] = raw_instance.get('ConnectionMode')
        instance_dict['region_id'] = raw_instance.get('RegionId')
        instance_dict['resource_group_id'] = raw_instance.get('ResourceGroupId')
        instance_dict['vswitch_id'] = raw_instance.get('VSwitchId')
        instance_dict['instance_network_type'] = raw_instance.get('InstanceNetworkType')
        instance_dict['db_instance_type'] = raw_instance.get('DBInstanceType')
        instance_dict['db_instance_status'] = raw_instance.get('DBInstanceStatus')
        instance_dict['zone_id'] = raw_instance.get('ZoneId')
        instance_dict['engine_version'] = raw_instance.get('EngineVersion')
        instance_dict['vpc_cloud_instance_id'] = raw_instance.get('VpcCloudInstanceId')
        instance_dict['pay_type'] = raw_instance.get('PayType')
        instance_dict['db_instance_class'] = raw_instance.get('DBInstanceClass')

        return instance_dict['id'], instance_dict




from ScoutSuite.providers.aliyun.resources.regions import Regions
from ScoutSuite.providers.aliyun.resources.vpc.vpcs import VPCs
from ScoutSuite.providers.aliyun.facade.base import AliyunFacade


class VPC(Regions):
    _children = [
        (VPCs, 'vpcs')
    ]

    def __init__(self, facade: AliyunFacade):
        super(VPC, self).__init__('vpc', facade)



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class Backups(Resources):
    def __init__(self, facade: GCPFacade, project_id: str, instance_name: str):
        super(Backups, self).__init__(facade)
        self.project_id = project_id
        self.instance_name = instance_name

    async def fetch_all(self):
        raw_backups = await self.facade.cloudsql.get_backups(self.project_id, self.instance_name)
        for raw_backup in raw_backups:
            if raw_backup['status'] == 'SUCCESSFUL':
                backup_id, backup = self._parse_backup(raw_backup)
                self[backup_id] = backup

    def _parse_backup(self, raw_backup):
        backup_dict = {}
        backup_dict['id'] = raw_backup['id']
        backup_dict['backup_url'] = raw_backup['selfLink'],
        backup_dict['creation_timestamp'] = raw_backup['endTime'],
        backup_dict['status'] = raw_backup['status'],
        backup_dict['type'] = raw_backup['type']
        return backup_dict['id'], backup_dict



from ScoutSuite.providers.aliyun.facade.base import AliyunFacade
from ScoutSuite.providers.aliyun.resources.base import AliyunCompositeResources
from ScoutSuite.providers.aliyun.resources.ram.groups import Groups
from ScoutSuite.providers.aliyun.resources.ram.password_policy import PasswordPolicy
from ScoutSuite.providers.aliyun.resources.ram.policies import Policies
from ScoutSuite.providers.aliyun.resources.ram.roles import Roles
from ScoutSuite.providers.aliyun.resources.ram.security_policy import SecurityPolicy
from ScoutSuite.providers.aliyun.resources.ram.users import Users


class RAM(AliyunCompositeResources):
    _children = [
        (Users, 'users'),
        (Groups, 'groups'),
        (Roles, 'roles'),
        (Policies, 'policies'),
        (PasswordPolicy, 'password_policy'),
        (SecurityPolicy, 'security_policy')
    ]

    def __init__(self, facade: AliyunFacade):
        super(RAM, self).__init__(facade)
        self.service = 'ram'

    async def fetch_all(self, **kwargs):
        await self._fetch_children(resource_parent=self)

        # We do not want the report to count the password policies as resources,
        # they aren't really resources.
        self['password_policy_count'] = 0
        self['security_policy_count'] = 0

        # TODO for each user check last login & API key usage for "last activity"

    async def finalize(self):
        self._match_users_and_groups()
        self._match_policies_and_entities()

    def _match_users_and_groups(self):
        """
        Parses the users and groups to match
        :return: None
        """
        for user in self['users']:
            self['users'][user]['groups'] = []
            for group in self['groups']:
                if any(u['name'] == user for u in self['groups'][group]['users']):
                    self['users'][user]['groups'].append(group)

    def _match_policies_and_entities(self):
        for policy in self['policies']:
            for user in self['users']:
                if not self['users'][user].get('policies'):
                    self['users'][user]['policies'] = []
                if self['users'][user]['name'] in self['policies'][policy]['entities'].get('users', []):
                    self['users'][user]['policies'].append(self['policies'][policy]['id'])
        for policy in self['policies']:
            for group in self['groups']:
                if not self['groups'][group].get('policies'):
                    self['groups'][group]['policies'] = []
                if self['groups'][group]['name'] in self['policies'][policy]['entities'].get('groups', []):
                    self['groups'][group]['policies'].append(self['policies'][policy]['id'])
        for policy in self['policies']:
            for role in self['roles']:
                if not self['roles'][role].get('policies'):
                    self['roles'][role]['policies'] = []
                if self['roles'][role]['name'] in self['policies'][policy]['entities'].get('roles', []):
                    self['roles'][role]['policies'].append(self['policies'][policy]['id'])



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .topics import Topics


class SNS(Regions):
    _children = [
        (Topics, 'topics')
    ]

    def __init__(self, facade: AWSFacade):
        super(SNS, self).__init__('sns', facade)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Subscriptions(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, topic_name: str):
        super(Subscriptions, self).__init__(facade)
        self.region = region
        self.topic_name = topic_name

    async def fetch_all(self):
        raw_subscriptions = await self.facade.sns.get_subscriptions(self.region, self.topic_name)
        self['protocol'] = {}
        self['subscriptions_count'] = 0
        for raw_subscription in raw_subscriptions:
            protocol, subscription = self._parse_subscription(raw_subscription)
            if protocol in self['protocol']:
                self['protocol'][protocol].append(subscription)
            else:
                self['protocol'][protocol] = [subscription]
            self['subscriptions_count'] += 1

    def _parse_subscription(self, raw_subscription):
        protocol = raw_subscription.pop('Protocol')
        return protocol, raw_subscription



from ScoutSuite.providers.aliyun.resources.base import AliyunResources


class Trails(AliyunResources):
    async def fetch_all(self):
        for raw_trail in await self.facade.actiontrail.get_trails():
            id, trail = self._parse_trails(raw_trail)
            self[id] = trail

    def _parse_trails(self, raw_trail):
        trail_dict = {}
        trail_dict['id'] = raw_trail.get('Name')
        trail_dict['name'] = raw_trail.get('Name')
        trail_dict['role_name'] = raw_trail.get('RoleName')
        trail_dict['home_region'] = raw_trail.get('HomeRegion')
        trail_dict['oss_bucket_name'] = raw_trail.get('OssBucketName')
        trail_dict['include_global_service_event'] = raw_trail.get('IncludeGlobalServiceEvent')
        trail_dict['status'] = raw_trail.get('Status')
        trail_dict['oss_key_prefix'] = raw_trail.get('OssKeyPrefix')
        trail_dict['region'] = raw_trail.get('Region')
        trail_dict['event_rw'] = raw_trail.get('EventRW')
        trail_dict['type'] = raw_trail.get('Type')
        trail_dict['sls_write_role_arn'] = raw_trail.get('SlsWriteRoleArn')
        trail_dict['sls_project_arn'] = raw_trail.get('SlsProjectArn')
        return trail_dict['id'], trail_dict




from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
import datetime


class Keys(Resources):
    def __init__(self, facade: GCPFacade, project_id: str, keyring_name: str, location: str):
        super(Keys, self).__init__(facade)
        self.project_id = project_id
        self.keyring_name = keyring_name
        self.location = location

    async def fetch_all(self):
        raw_keys = await self.facade.kms.list_keys(self.project_id, self.location, self.keyring_name)
        for raw_key in raw_keys:
            key_id, key = self._parse_key(raw_key)
            self[key_id] = key

    def _parse_key(self, raw_key):

        key_dict = {}

        key_dict['id'] = raw_key['name'].split('/')[-1]
        key_dict['state'] = raw_key['primary']['state']
        key_dict['creation_datetime'] = raw_key['primary']['createTime']
        key_dict['protection_level'] = raw_key['primary']['protectionLevel']
        key_dict['algorithm'] = raw_key['primary']['algorithm']
        key_dict['next_rotation_datetime'] = raw_key.get('nextRotationTime', None)
        key_dict['purpose'] = raw_key['purpose']
        key_dict['rotation_period'] = raw_key.get('rotationPeriod', None)

        return key_dict['id'], key_dict



from abc import ABCMeta, abstractmethod


class AuthenticationStrategy(metaclass=ABCMeta):
    """
    This class represents an authentication strategy.
    """

    @abstractmethod
    def authenticate(self, **kwargs):
        """
        Given parameters, this authenticates the user to a provider and returns a credentials object.
        """
        raise NotImplementedError()


class AuthenticationException(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors



from ScoutSuite.providers.aliyun.resources.base import AliyunResources
from ScoutSuite.providers.aliyun.facade.base import AliyunFacade


class ApiKeys(AliyunResources):
    def __init__(self, facade: AliyunFacade, user):
        super(ApiKeys, self).__init__(facade)
        self.user = user

    async def fetch_all(self):
        for raw_user_api_key in await self.facade.ram.get_user_api_keys(username=self.user['name']):
            id, api_key = await self._parse_api_key(raw_user_api_key)
            self[id] = api_key

    async def _parse_api_key(self, raw_api_key):
        api_key = {}
        api_key['id'] = raw_api_key['AccessKeyId']
        api_key['creation_datetime'] = raw_api_key['CreateDate']
        api_key['status'] = raw_api_key['Status']

        last_usage = await self.facade.ram.get_user_api_key_last_usage(self.user['name'], api_key['id'])
        api_key['last_usage_datetime'] = last_usage if last_usage != 'N/A' else None

        return api_key['id'], api_key



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Clusters(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super(Clusters, self).__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_clusters = await self.facade.redshift.get_clusters(self.region, self.vpc)
        for raw_cluster in raw_clusters:
            id, cluster = self._parse_cluster(raw_cluster)
            self[id] = cluster

    def _parse_cluster(self, raw_cluster):
        name = raw_cluster.pop('ClusterIdentifier')
        raw_cluster['name'] = name

        return name, raw_cluster



import asyncio
import base64
import boto3

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import get_and_set_concurrently
from ScoutSuite.providers.utils import run_concurrently


class EC2Facade(AWSBaseFacade):
    regional_flow_logs_cache_locks = {}
    flow_logs_cache = {}

    def __init__(self, session: boto3.session.Session, owner_id: str):
        self.owner_id = owner_id

        super(EC2Facade, self).__init__(session)

    async def get_instance_user_data(self, region: str, instance_id: str):
        ec2_client = AWSFacadeUtils.get_client('ec2', self.session, region)
        try:
            user_data_response = await run_concurrently(
                lambda: ec2_client.describe_instance_attribute(Attribute='userData', InstanceId=instance_id))
        except Exception as e:
            print_exception(
                'Failed to describe EC2 instance attributes: {}'.format(e))
            return None
        else:
            if 'Value' not in user_data_response['UserData'].keys():
                return None
            return base64.b64decode(user_data_response['UserData']['Value']).decode('utf-8')

    async def get_instances(self, region: str, vpc: str):
        filters = [{'Name': 'vpc-id', 'Values': [vpc]}]
        try:
            reservations = \
                await AWSFacadeUtils.get_all_pages(
                    'ec2', region, self.session, 'describe_instances', 'Reservations', Filters=filters)

            instances = []
            for reservation in reservations:
                for instance in reservation['Instances']:
                    instance['ReservationId'] = reservation['ReservationId']
                    instances.append(instance)

            return instances
        except Exception as e:
            print_exception('Failed to describe EC2 instances: {}'.format(e))
            return []

    async def get_security_groups(self, region: str, vpc: str):
        filters = [{'Name': 'vpc-id', 'Values': [vpc]}]
        try:
            return await AWSFacadeUtils.get_all_pages(
                'ec2', region, self.session, 'describe_security_groups', 'SecurityGroups', Filters=filters)
        except Exception as e:
            print_exception('Failed to describe EC2 security groups: {}'.format(e))
            return []

    async def get_vpcs(self, region: str):
        ec2_client = AWSFacadeUtils.get_client('ec2', self.session, region)
        try:
            return await run_concurrently(lambda: ec2_client.describe_vpcs()['Vpcs'])
        except Exception as e:
            print_exception('Failed to describe EC2 VPC: {}'.format(e))
            return []

    async def get_images(self, region: str):
        filters = [{'Name': 'owner-id', 'Values': [self.owner_id]}]
        client = AWSFacadeUtils.get_client('ec2', self.session, region)
        try:
            return await run_concurrently(lambda: client.describe_images(Filters=filters)['Images'])
        except Exception as e:
            print_exception('Failed to get EC2 images: {}'.format(e))
            return []

    async def get_network_interfaces(self, region: str, vpc: str):
        filters = [{'Name': 'vpc-id', 'Values': [vpc]}]
        try:
            return await AWSFacadeUtils.get_all_pages(
                'ec2', region, self.session, 'describe_network_interfaces', 'NetworkInterfaces', Filters=filters)
        except Exception as e:
            print_exception('Failed to get EC2 network interfaces: {}'.format(e))
            return []

    async def get_volumes(self, region: str):
        try:
            volumes = await AWSFacadeUtils.get_all_pages('ec2', region, self.session, 'describe_volumes', 'Volumes')
            await get_and_set_concurrently([self._get_and_set_key_manager], volumes, region=region)
            return volumes
        except Exception as e:
            print_exception('Failed to get EC2 volumes: {}'.format(e))
            return []

    async def _get_and_set_key_manager(self, volume: {}, region: str):
        kms_client = AWSFacadeUtils.get_client('kms', self.session, region)
        if 'KmsKeyId' in volume:
            key_id = volume['KmsKeyId']
            try:
                volume['KeyManager'] = await run_concurrently(
                    lambda: kms_client.describe_key(KeyId=key_id)['KeyMetadata']['KeyManager'])
            except Exception as e:
                print_exception('Failed to describe KMS key: {}'.format(e))
                volume['KeyManager'] = None
        else:
            volume['KeyManager'] = None

    async def get_snapshots(self, region: str):
        filters = [{'Name': 'owner-id', 'Values': [self.owner_id]}]

        try:
            snapshots = await AWSFacadeUtils.get_all_pages(
                'ec2', region, self.session, 'describe_snapshots', 'Snapshots', Filters=filters)
        except Exception as e:
            print_exception('Failed to get snapshots: {}'.format(e))
            snapshots = []
        else:
            await get_and_set_concurrently([self._get_and_set_snapshot_attributes], snapshots, region=region)
        finally:
            return snapshots

    async def _get_and_set_snapshot_attributes(self, snapshot: {}, region: str):
        ec2_client = AWSFacadeUtils.get_client('ec2', self.session, region)
        try:
            snapshot['CreateVolumePermissions'] = await run_concurrently(lambda: ec2_client.describe_snapshot_attribute(
                Attribute='createVolumePermission',
                SnapshotId=snapshot['SnapshotId'])['CreateVolumePermissions'])
        except Exception as e:
            print_exception(
                'Failed to describe EC2 snapshot attributes: {}'.format(e))

    async def get_network_acls(self, region: str, vpc: str):
        filters = [{'Name': 'vpc-id', 'Values': [vpc]}]
        try:
            return await AWSFacadeUtils.get_all_pages(
                'ec2', region, self.session, 'describe_network_acls', 'NetworkAcls', Filters=filters)
        except Exception as e:
            print_exception('Failed to get EC2 network ACLs: {}'.format(e))
            return []

    async def get_flow_logs(self, region: str):
        try:
            await self.cache_flow_logs(region)
            return self.flow_logs_cache[region]
        except Exception as e:
            print_exception('Failed to get EC2 flow logs: {}'.format(e))
            return []

    async def cache_flow_logs(self, region: str):
        async with self.regional_flow_logs_cache_locks.setdefault(region, asyncio.Lock()):
            if region in self.flow_logs_cache:
                return

            self.flow_logs_cache[region] = \
                await AWSFacadeUtils.get_all_pages('ec2', region, self.session, 'describe_flow_logs', 'FlowLogs')

    async def get_subnets(self, region: str, vpc: str):
        ec2_client = AWSFacadeUtils.get_client('ec2', self.session, region)
        filters = [{'Name': 'vpc-id', 'Values': [vpc]}]
        try:
            subnets = await run_concurrently(lambda: ec2_client.describe_subnets(Filters=filters)['Subnets'])
        except Exception as e:
            print_exception('Failed to describe EC2 subnets: {}'.format(e))
            return None
        else:
            await get_and_set_concurrently([self._get_and_set_subnet_flow_logs], subnets, region=region)
            return subnets

    async def _get_and_set_subnet_flow_logs(self, subnet: {}, region: str):
        await self.cache_flow_logs(region)
        subnet['flow_logs'] = \
            [flow_log for flow_log in self.flow_logs_cache[region]
             if flow_log['ResourceId'] == subnet['SubnetId'] or flow_log['ResourceId'] == subnet['VpcId']]

    async def get_peering_connections(self, region):
        try:
            peering_connections = await AWSFacadeUtils.get_all_pages('ec2', region, self.session, 'describe_vpc_peering_connections', 'VpcPeeringConnections')
            return peering_connections
        except Exception as e:
            print_exception('Failed to get peering connections: {}'.format(e))
            return []



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.gcp.resources.base import GCPCompositeResources
from ScoutSuite.providers.gcp.resources.cloudsql.backups import Backups
from ScoutSuite.providers.gcp.resources.cloudsql.users import Users
from ScoutSuite.providers.utils import get_non_provider_id


class DatabaseInstances(GCPCompositeResources):
    _children = [
        (Backups, 'backups'),
        (Users, 'users')
    ]

    def __init__(self, facade: GCPFacade, project_id: str):
        super(DatabaseInstances, self).__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_instances = await self.facade.cloudsql.get_database_instances(self.project_id)
        for raw_instance in raw_instances:
            instance_id, instance = self._parse_instance(raw_instance)
            self[instance_id] = instance
        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={instance_id: {'project_id': self.project_id, 'instance_name': instance['name']}
                    for instance_id, instance in self.items()})
        self._set_last_backup_timestamps(self.items())

    def _parse_instance(self, raw_instance):
        instance_dict = {}

        instance_dict['id'] = get_non_provider_id(raw_instance['name'])
        instance_dict['name'] = raw_instance['name']
        instance_dict['project_id'] = raw_instance['project']
        instance_dict['automatic_backup_enabled'] = raw_instance['settings']['backupConfiguration']['enabled']
        instance_dict['database_version'] = raw_instance['databaseVersion']
        instance_dict['log_enabled'] = self._is_log_enabled(raw_instance)
        instance_dict['ssl_required'] = self._is_ssl_required(raw_instance)
        instance_dict['authorized_networks'] = raw_instance['settings']['ipConfiguration']['authorizedNetworks']

        # check if is or has a failover replica
        instance_dict['has_failover_replica'] = raw_instance.get('failoverReplica', []) != []
        instance_dict['is_failover_replica'] = raw_instance.get('masterInstanceName', '') != ''

        # network interfaces
        instance_dict['public_ip'] = None
        instance_dict['private_ip'] = None
        for address in raw_instance.get('ipAddresses', []):
            if address['type'] == 'PRIMARY':
                instance_dict['public_ip'] = address['ipAddress']
            elif address['type'] == 'PRIVATE':
                instance_dict['private_ip'] = address['ipAddress']
            else:
                print_exception('Unknown Cloud SQL instance IP address type: {}'.format(address['type']))

        return instance_dict['id'], instance_dict

    def _is_log_enabled(self, raw_instance):
        return raw_instance['settings']['backupConfiguration'].get('binaryLogEnabled')

    def _is_ssl_required(self, raw_instance):
        return raw_instance['settings']['ipConfiguration'].get('requireSsl')

    def _set_last_backup_timestamps(self, instances):
        for instance_id, _ in instances:
            self[instance_id]['last_backup_timestamp'] = self._get_last_backup_timestamp(
                self[instance_id]['backups'])

    def _get_last_backup_timestamp(self, backups):
        if not backups:
            return 'N/A'
        last_backup_id = max(backups.keys(), key=(
            lambda k: backups[k]['creation_timestamp']))
        return backups[last_backup_id]['creation_timestamp']



import json

from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class IdentityPolicies(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, identity_name: str):
        super(IdentityPolicies, self).__init__(facade)
        self.region = region
        self.identity_name = identity_name

    async def fetch_all(self):
        raw_policies = await self.facade.ses.get_identity_policies(self.region, self.identity_name)
        for policy_name, raw_policy in raw_policies.items():
            self[policy_name] = json.loads(raw_policy)



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import get_name


class Volumes(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(Volumes, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_volumes = await self.facade.ec2.get_volumes(self.region)
        for raw_volume in raw_volumes:
            name, resource = self._parse_volume(raw_volume)
            self[name] = resource

    def _parse_volume(self, raw_volume):
        raw_volume['id'] = raw_volume.pop('VolumeId')
        raw_volume['name'] = get_name(raw_volume, raw_volume, 'id')
        return raw_volume['id'], raw_volume



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class CredentialReports(AWSResources):
    async def fetch_all(self):
        raw_credential_reports = await self.facade.iam.get_credential_reports()
        for raw_credential_report in raw_credential_reports:
            name, resource = self._parse_credential_reports(raw_credential_report)
            self[name] = resource

    def _parse_credential_reports(self, raw_credential_report):
        user_id = raw_credential_report['user']
        raw_credential_report['name'] = user_id
        raw_credential_report['id'] = user_id
        raw_credential_report['password_last_used'] = self._sanitize_date(raw_credential_report['password_last_used'])
        raw_credential_report['access_key_1_last_used_date'] =\
            self._sanitize_date(raw_credential_report['access_key_1_last_used_date'])
        raw_credential_report['access_key_2_last_used_date'] =\
            self._sanitize_date(raw_credential_report['access_key_2_last_used_date'])
        raw_credential_report['last_used'] = self._compute_last_used(raw_credential_report)
        return get_non_provider_id(user_id), raw_credential_report

    @staticmethod
    def _sanitize_date(date):
        """
        Returns the date if it is not equal to 'N/A' or 'no_information', else returns None
        """
        return date if date != 'no_information' and date != 'N/A' else None

    @staticmethod
    def _compute_last_used(credential_report):
        dates = [credential_report['password_last_used'],
                 credential_report['access_key_1_last_used_date'],
                 credential_report['access_key_2_last_used_date']]

        dates = [date for date in dates if date is not None]
        return max(dates) if len(dates) > 0 else None



from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class NetworkInterfaces(AzureResources):
    async def fetch_all(self):
        for raw_network_interface in await self.facade.network.get_network_interfaces():
            id, network_interface = self._parse_network_interface(raw_network_interface)
            self[id] = network_interface

    def _parse_network_interface(self, raw_network_interface):
        network_interface_dict = {}
        network_interface_dict['id'] = get_non_provider_id(raw_network_interface.id)
        network_interface_dict['enable_accelerated_networking'] = raw_network_interface.enable_accelerated_networking
        network_interface_dict['virtual_machine'] = \
            get_non_provider_id(raw_network_interface.virtual_machine.id.lower()) if \
                raw_network_interface.virtual_machine else None
        network_interface_dict['name'] = raw_network_interface.name
        network_interface_dict['tags'] = raw_network_interface.tags
        network_interface_dict['interface_endpoint'] = raw_network_interface.interface_endpoint if \
            hasattr(raw_network_interface, 'interface_endpoint') else None
        network_interface_dict['primary'] = raw_network_interface.primary
        network_interface_dict['etag'] = raw_network_interface.etag
        network_interface_dict['additional_properties'] = raw_network_interface.additional_properties
        network_interface_dict['location'] = raw_network_interface.location
        network_interface_dict['mac_address'] = raw_network_interface.mac_address
        network_interface_dict['provisioning_state'] = raw_network_interface.provisioning_state
        network_interface_dict['resource_guid'] = raw_network_interface.resource_guid
        network_interface_dict['enable_ip_forwarding'] = raw_network_interface.enable_ip_forwarding
        network_interface_dict['type'] = raw_network_interface.type
        network_interface_dict['network_security_group'] = get_non_provider_id(raw_network_interface.network_security_group.id)

        # TODO process and display the below
        network_interface_dict['hosted_workloads'] = raw_network_interface.hosted_workloads
        network_interface_dict['tap_configurations'] = raw_network_interface.tap_configurations
        network_interface_dict['dns_settings'] = raw_network_interface.dns_settings

        ip_configuration = raw_network_interface.ip_configurations[0]  # TODO is this always an array of 1?
        network_interface_dict['ip_configuration'] = {}
        network_interface_dict['ip_configuration']['id'] = ip_configuration.id
        network_interface_dict['ip_configuration']['additional_properties'] = ip_configuration.additional_properties
        network_interface_dict['ip_configuration']['virtual_network_taps'] = ip_configuration.virtual_network_taps
        network_interface_dict['ip_configuration'][
            'application_gateway_backend_address_pools'] = ip_configuration.application_gateway_backend_address_pools
        network_interface_dict['ip_configuration'][
            'load_balancer_backend_address_pools'] = ip_configuration.load_balancer_backend_address_pools
        network_interface_dict['ip_configuration'][
            'load_balancer_inbound_nat_rules'] = ip_configuration.load_balancer_inbound_nat_rules
        network_interface_dict['ip_configuration']['private_ip_address'] = ip_configuration.private_ip_address
        network_interface_dict['ip_configuration'][
            'private_ip_allocation_method'] = ip_configuration.private_ip_allocation_method
        network_interface_dict['ip_configuration'][
            'private_ip_address_version'] = ip_configuration.private_ip_address_version
        network_interface_dict['ip_configuration']['subnet'] = {'id': get_non_provider_id(ip_configuration.subnet.id)}
        network_interface_dict['ip_configuration']['primary'] = ip_configuration.primary
        network_interface_dict['ip_configuration']['public_ip_address'] = ip_configuration.public_ip_address
        network_interface_dict['ip_configuration'][
            'application_security_groups'] = ip_configuration.application_security_groups
        network_interface_dict['ip_configuration']['provisioning_state'] = ip_configuration.provisioning_state
        network_interface_dict['ip_configuration']['name'] = ip_configuration.name
        network_interface_dict['ip_configuration']['etag'] = ip_configuration.etag

        return network_interface_dict['id'], network_interface_dict




from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class KeyVaults(AzureResources):
    async def fetch_all(self):
        self['vaults'] = {}
        for raw_vault in await self.facade.keyvault.get_key_vaults():
            id, vault = self._parse_key_vault(raw_vault)
            self['vaults'][id] = vault

        self['vaults_count'] = len(self['vaults'])

    def _parse_key_vault(self, raw_vault):
        vault = {}
        vault['id'] = get_non_provider_id(raw_vault.id)
        vault['name'] = raw_vault.name
        vault['public_access_allowed'] = self._is_public_access_allowed(raw_vault)

        return vault['id'], vault

    def _is_public_access_allowed(self, raw_vault):
        return raw_vault.properties.network_acls is None



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class Bindings(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super(Bindings, self).__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_bindings = await self.facade.cloudresourcemanager.get_bindings(self.project_id)
        for raw_binding in raw_bindings:
            binding_id, binding = self._parse_binding(raw_binding)
            self[binding_id] = binding

    def _parse_binding(self, raw_binding):
        binding_dict = {}
        binding_dict['id'] = get_non_provider_id(raw_binding['role'])
        binding_dict['name'] = raw_binding['role'].split('/')[-1]
        binding_dict['members'] = self._parse_members(raw_binding)
        return binding_dict['id'], binding_dict

    def _parse_members(self, raw_binding):
        members_dict = {'users': [], 'groups': [], 'service_accounts': []}
        
        if 'members' not in raw_binding:
            return members_dict

        type_map = { 
            'user': 'users', 
            'group': 'groups', 
            'serviceAccount': 'service_accounts'
        }
        
        # We want to group the members by type, so we need to parse their type and entity.
        # The members are given as strings with the format <member_type>:<member_entity>
        # See the GCP Resource Manager API reference for more info:
        # https://cloud.google.com/resource-manager/reference/rest/Shared.Types/Binding 
        for member in raw_binding['members']:
            member_type, entity = member.split(':')[:2]
            if member_type in type_map:
                members_dict[type_map[member_type]].append(entity)
            else:
                print_exception('Type %s not handled' % member_type)
        
        return members_dict



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class ParameterGroups(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(ParameterGroups, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_parameter_groups = await self.facade.elasticache.get_parameter_groups(self.region)
        for raw_parameter_group in raw_parameter_groups:
            name, resource = self._parse_parameter_group(raw_parameter_group)
            self[name] = resource

    def _parse_parameter_group(self, raw_parameter_group):
        raw_parameter_group['name'] = raw_parameter_group.pop('CacheParameterGroupName')
        return raw_parameter_group['name'], raw_parameter_group



import abc

from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from ScoutSuite.providers.aws.facade.base import AWSFacade


class Regions(AWSCompositeResources, metaclass=abc.ABCMeta):
    def __init__(self, service: str, facade: AWSFacade):
        super(Regions, self).__init__(facade)
        self.service = service

    async def fetch_all(self, regions=None, excluded_regions=None, partition_name='aws', **kwargs):
        self['regions'] = {}
        for region in await self.facade.build_region_list(self.service, regions, excluded_regions, partition_name):
            self['regions'][region] = {
                'id': region,
                'region': region,
                'name': region
            }

        await self._fetch_children_of_all_resources(
            resources=self['regions'],
            scopes={region: {'region': region} for region in self['regions']}
        )

        self._set_counts()

    def _set_counts(self):
        self['regions_count'] = len(self['regions'])
        for _, key in self._children:
            # VPCs should not be counted as resources. They exist whether you have resources or not, so
            # counting them would make the report confusing.
            if key == 'vpcs':
                continue
                
            self[key + '_count'] = sum([region[key + '_count'] for region in self['regions'].values()])



import re

from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Stacks(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(Stacks, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_stacks = await self.facade.cloudformation.get_stacks(self.region)
        for raw_stack in raw_stacks:
            name, stack = self._parse_stack(raw_stack)
            self[name] = stack

    def _parse_stack(self, raw_stack):
        raw_stack['id'] = raw_stack.pop('StackId')
        raw_stack['name'] = raw_stack.pop('StackName')
        raw_stack['drifted'] = raw_stack.pop('DriftInformation')[
                                   'StackDriftStatus'] == 'DRIFTED'
        raw_stack['termination_protection'] = raw_stack['EnableTerminationProtection']

        template = raw_stack.pop('template')
        raw_stack['deletion_policy'] = self.has_deletion_policy(template)

        if hasattr(template, 'keys'):
            for group in template.keys():
                if 'DeletionPolicy' in template[group]:
                    raw_stack['deletion_policy'] = template[group]['DeletionPolicy']
                    break

        return raw_stack['name'], raw_stack

    @staticmethod
    def has_deletion_policy(template):
        """
        Return region to be used for global calls such as list bucket and get bucket location
        :param template:                    The api response containing the stack's template
        :return:
        """
        has_dp = True
        # If a ressource is found to not have a deletion policy or have it to delete, the boolean is switched to
        # false to indicate that the ressource will be deleted once the stack is deleted
        if isinstance(template, dict):
            template = template['Resources']
            for group in template.keys():
                if 'DeletionPolicy' in template[group]:
                    if template[group]['DeletionPolicy'] == 'Delete':
                        has_dp = False
                else:
                    has_dp = False
        if isinstance(template, str):
            if re.match(r'\"DeletionPolicy\"\s*:\s*\"Delete\"', template):
                has_dp = False
            elif not re.match(r'\"DeletionPolicy\"', template):
                has_dp = False
        return has_dp



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.kms.keyrings import KeyRings


class KMS(Projects):
    _children = [
        (KeyRings, 'keyrings')
    ]



from ScoutSuite.providers.aliyun.authentication_strategy import AliyunCredentials
from ScoutSuite.providers.aliyun.facade.utils import get_response
from ScoutSuite.providers.aliyun.utils import get_client

from ScoutSuite.core.console import print_exception
from aliyunsdkram.request.v20150501 import \
    ListUsersRequest, GetUserRequest, \
    GetUserMFAInfoRequest, \
    ListAccessKeysRequest, GetAccessKeyLastUsedRequest, \
    GetPasswordPolicyRequest, GetSecurityPreferenceRequest, \
    ListGroupsRequest, ListUsersForGroupRequest, \
    ListRolesRequest, \
    ListPoliciesRequest, GetPolicyVersionRequest, ListEntitiesForPolicyRequest


class RAMFacade:
    def __init__(self, credentials: AliyunCredentials):
        self._credentials = credentials
        self._client = get_client(credentials=self._credentials)

    async def get_users(self):
        """
        Get all users

        :return: a list of all users
        """
        response = await get_response(client=self._client,
                                      request=ListUsersRequest.ListUsersRequest())
        if response:
            return response['Users']['User']
        else:
            return []

    async def get_user_details(self, username):
        """
        Get additional details for a user

        :param username: The username of the user
        :return: a dict with the user's details
        """
        request = GetUserRequest.GetUserRequest()
        request.set_UserName(username)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['User']
        else:
            return []

    async def get_user_api_keys(self, username):
        """
        Get API keys for a user

        :param username: The username of the user
        :return: the list of API keys for that user
        """
        request = ListAccessKeysRequest.ListAccessKeysRequest()
        request.set_UserName(username)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['AccessKeys']['AccessKey']
        else:
            return []

    async def get_user_api_key_last_usage(self, username, key_id):
        """
        Get last usage date for an API key

        :param username: The username of the user
        :param key_id: The API key id
        :return: the list of API keys for that user
        """
        request = GetAccessKeyLastUsedRequest.GetAccessKeyLastUsedRequest()
        request.set_UserName(username)
        request.set_UserAccessKeyId(key_id)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['AccessKeyLastUsed']['LastUsedDate']
        else:
            return []

    async def get_user_mfa_status(self, username):
        """
        Check if user has MFA configured

        :param username: The username of the user
        :return: status and MFA serial number
        """
        request = GetUserMFAInfoRequest.GetUserMFAInfoRequest()
        request.set_UserName(username)
        try:
            response = await get_response(client=self._client,
                                          request=request)
        except Exception as e:
            # TODO can't seem to differenciate between a user that has MFA disabled
            # and a user that has MFA enabled but not configured
            if e.error_code == 'EntityNotExist.User.MFADevice':
                # ignore, MFA is not configured
                return False, None
            else:
                print_exception('Unable to get MFA status for user {}: {}'.format(username,
                                                                                  e))
                return False, None
        else:
            return True, response['MFADevice']['SerialNumber']

    async def get_password_policy(self):
        """
        Get the account's password policy

        :return: the password policy
        """
        request = GetPasswordPolicyRequest.GetPasswordPolicyRequest()
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['PasswordPolicy']
        else:
            return []

    async def get_security_policy(self):
        """
        Get the account's security policy

        :return: the security policy
        """
        request = GetSecurityPreferenceRequest.GetSecurityPreferenceRequest()
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['SecurityPreference']
        else:
            return []

    async def get_groups(self):
        """
        Get all groups

        :return: a list of all groups
        """
        response = await get_response(client=self._client,
                                      request=ListGroupsRequest.ListGroupsRequest())
        if response:
            return response['Groups']['Group']
        else:
            return []

    async def get_group_users(self, group_name):
        """
        Get all users in a group

        :return: a list of users in groups
        """
        request = ListUsersForGroupRequest.ListUsersForGroupRequest()
        request.set_GroupName(group_name)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['Users']['User']
        else:
            return []

    async def get_roles(self):
        """
        Get all roles

        :return: a list of all roles
        """
        response = await get_response(client=self._client,
                                      request=ListRolesRequest.ListRolesRequest())
        if response:
            return response['Roles']['Role']
        else:
            return []

    async def get_policies(self):
        """
        Get all custom policies

        :return: a list of all custom policies
        """
        response = await get_response(client=self._client,
                                      request=ListPoliciesRequest.ListPoliciesRequest())
        if response:
            return response['Policies']['Policy']
        else:
            return []

    async def get_policy_version(self, name, type, version):
        """
        Get all policies

        :return: a list of all policies
        """
        request = GetPolicyVersionRequest.GetPolicyVersionRequest()
        request.set_PolicyName(name)
        request.set_PolicyType(type)
        request.set_VersionId(version)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            return response['PolicyVersion']
        else:
            return []

    async def get_policy_entities(self, name, type):
        """
        Get all entities for a policy

        :return: a dict of all policy entities
        """
        request = ListEntitiesForPolicyRequest.ListEntitiesForPolicyRequest()
        request.set_PolicyName(name)
        request.set_PolicyType(type)
        response = await get_response(client=self._client,
                                      request=request)
        if response:
            response.pop('RequestId')
            return response
        else:
            return []



from ScoutSuite.providers.gcp.resources.base import GCPCompositeResources


class Projects(GCPCompositeResources):

    """This class represents a collection of GCP Resources that are grouped by project. 
    Classes extending Projects should implement the method _fetch_children() with a project ID as paramater.
    The children resources will be stored with the following structure {<projects>: {<project_id>: {<child_name>: {<child_id>: <child_instance>}}}}.
    """

    async def fetch_all(self):
        """This method fetches all the GCP projects that can be accessed with the given run configuration.
        It then fetches all the children defined in _children and groups them by project.
        """

        raw_projects = await self.facade.get_projects()
        self['projects'] = {raw_project['projectId']: {}
                            for raw_project in raw_projects}
        await self._fetch_children_of_all_resources(
            resources=self['projects'],
            scopes={project_id: {'project_id': project_id} for project_id in self['projects']})
        self._set_counts()

    def _set_counts(self):
        for _, child_name in self._children:
            self[child_name + '_count'] = sum([project[child_name + '_count']
                                               for project in self['projects'].values()])



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class ParameterGroups(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(ParameterGroups, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_parameter_groups = await self.facade.rds.get_parameter_groups(self.region)
        for raw_parameter_group in raw_parameter_groups:
            name, resource = self._parse_parameter_group(raw_parameter_group)
            self[name] = resource

    def _parse_parameter_group(self, raw_parameter_group):
        raw_parameter_group['arn'] = raw_parameter_group.pop('DBParameterGroupArn')
        raw_parameter_group['name'] = raw_parameter_group.pop('DBParameterGroupName')
        raw_parameter_group['parameters'] = raw_parameter_group.pop('Parameters')
        parameter_group_id = get_non_provider_id(raw_parameter_group['name'])
        return parameter_group_id, raw_parameter_group



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class Buckets(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super(Buckets, self).__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_buckets = await self.facade.cloudstorage.get_buckets(self.project_id)
        for raw_bucket in raw_buckets:
            bucket_id, bucket = self._parse_bucket(raw_bucket)
            self[bucket_id] = bucket

    def _parse_bucket(self, raw_bucket):
        bucket_dict = {}
        bucket_dict['id'] = get_non_provider_id(raw_bucket.id)
        bucket_dict['name'] = raw_bucket.name
        bucket_dict['project_id'] = self.project_id
        bucket_dict['project_number'] = raw_bucket.project_number
        bucket_dict['creation_date'] = raw_bucket.time_created
        bucket_dict['location'] = raw_bucket.location
        bucket_dict['storage_class'] = raw_bucket.storage_class.lower()
        bucket_dict['versioning_status_enabled'] = raw_bucket.versioning_enabled
        bucket_dict['logging_enabled'] = raw_bucket.logging is not None
        bucket_dict['acl_configuration'] = self._get_cloudstorage_bucket_acl(raw_bucket)
        return bucket_dict['id'], bucket_dict

    def _get_cloudstorage_bucket_acl(self, raw_bucket):
        bucket_acls = raw_bucket.iam_policy
        acl_config = {}
        if bucket_acls:
            for role in bucket_acls._bindings:
                for member in bucket_acls[role]:
                    if member.split(':')[0] not in ['projectEditor', 'projectViewer', 'projectOwner']:
                        if member not in acl_config:
                            acl_config[member] = [role]
                        else:
                            acl_config[member].append(role)
        return acl_config



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from ScoutSuite.providers.utils import get_non_provider_id
from .listeners import Listeners


class LoadBalancers(AWSCompositeResources):
    _children = [
        (Listeners, 'listeners')
    ]

    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super(LoadBalancers, self).__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_load_balancers = await self.facade.elbv2.get_load_balancers(self.region, self.vpc)
        for raw_load_balancer in raw_load_balancers:
            id, load_balancer = self._parse_load_balancer(raw_load_balancer)
            self[id] = load_balancer

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={load_balancer_id: {'region': self.region, 'load_balancer_arn': load_balancer['arn']}
                    for (load_balancer_id, load_balancer) in self.items()}
        )

    def _parse_load_balancer(self, load_balancer):
        load_balancer['arn'] = load_balancer.pop('LoadBalancerArn')
        load_balancer['name'] = load_balancer.pop('LoadBalancerName')
        load_balancer['security_groups'] = []

        if 'SecurityGroups' in load_balancer:
            for sg in load_balancer['SecurityGroups']:
                load_balancer['security_groups'].append({'GroupId': sg})
            load_balancer.pop('SecurityGroups')

        return get_non_provider_id(load_balancer['name']), load_balancer



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions
from .stacks import Stacks


class CloudFormation(Regions):
    _children = [
        (Stacks, 'stacks')
    ]

    def __init__(self, facade: AWSFacade):
        super(CloudFormation, self).__init__('cloudformation', facade)



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils


class Route53Facade(AWSBaseFacade):
    async def get_domains(self, region):
        try:
            return await AWSFacadeUtils.get_all_pages('route53domains', region, self.session,
                                                      'list_domains', 'Domains')
        except Exception as e:
            print_exception('Failed to get Route53 domains: {}'.format(e))
            return []

    async def get_hosted_zones(self):
        try:
            return await AWSFacadeUtils.get_all_pages('route53', None, self.session,
                                                      'list_hosted_zones', 'HostedZones')
        except Exception as e:
            print_exception('Failed to get Route53 hosted zones: {}'.format(e))

    async def get_resource_records(self, hosted_zone_id):
        try:
            return await AWSFacadeUtils.get_all_pages('route53', None, self.session,
                                                      'list_resource_record_sets', 'ResourceRecordSets',
                                                      HostedZoneId=hosted_zone_id)
        except Exception as e:
            print_exception('Failed to get Route53 resource records: {}'.format(e))



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .domains import Domains
from .hosted_zones import HostedZones


class Route53(Regions):
    _children = [
        (Domains, 'domains'),
        (HostedZones, 'hosted_zones')
    ]

    def __init__(self, facade: AWSFacade):
        super(Route53, self).__init__('route53domains', facade)



from ScoutSuite.providers.aws.resources.ec2.ami import AmazonMachineImages
from ScoutSuite.providers.aws.resources.ec2.snapshots import Snapshots
from ScoutSuite.providers.aws.resources.ec2.volumes import Volumes
from ScoutSuite.providers.aws.resources.ec2.vpcs import Ec2Vpcs
from ScoutSuite.providers.aws.resources.regions import Regions


class EC2(Regions):
    _children = [
        (Ec2Vpcs, 'vpcs'),
        (AmazonMachineImages, 'images'),
        (Snapshots, 'snapshots'),
        (Volumes, 'volumes')
    ]

    def __init__(self, facade):
        super(EC2, self).__init__('ec2', facade)

    async def fetch_all(self, regions=None, excluded_regions=None, partition_name='aws', **kwargs):
        await super(EC2, self).fetch_all(regions, excluded_regions, partition_name)

        for region in self['regions']:
            self['regions'][region]['instances_count'] =\
                sum([len(vpc['instances']) for vpc in self['regions'][region]['vpcs'].values()])
            self['regions'][region]['security_groups_count'] =\
                sum([len(vpc['security_groups']) for vpc in self['regions'][region]['vpcs'].values()])
            self['regions'][region]['network_interfaces_count'] =\
                sum([len(vpc['network_interfaces']) for vpc in self['regions'][region]['vpcs'].values()])
        
        self['instances_count'] =\
            sum([region['instances_count'] for region in self['regions'].values()])
        self['security_groups_count'] =\
            sum([region['security_groups_count'] for region in self['regions'].values()])
        self['network_interfaces_count'] =\
            sum([region['network_interfaces_count'] for region in self['regions'].values()])



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import get_name
from ScoutSuite.providers.aws.utils import get_keys
import re


class EC2Instances(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super(EC2Instances, self).__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_instances = await self.facade.ec2.get_instances(self.region, self.vpc)
        for raw_instance in raw_instances:
            name, resource = await self._parse_instance(raw_instance)
            self[name] = resource

    async def _parse_instance(self, raw_instance):
        instance = {}
        id = raw_instance['InstanceId']
        instance['id'] = id
        instance['reservation_id'] = raw_instance['ReservationId']
        instance['monitoring_enabled'] = raw_instance['Monitoring']['State'] == 'enabled'
        instance['user_data'] = await self.facade.ec2.get_instance_user_data(self.region, id)
        instance['user_data_secrets'] = self._identify_user_data_secrets(instance['user_data'])

        get_name(raw_instance, instance, 'InstanceId')
        get_keys(raw_instance, instance,
                 ['KeyName', 'LaunchTime', 'InstanceType', 'State', 'IamInstanceProfile', 'SubnetId'])

        instance['network_interfaces'] = {}
        for eni in raw_instance['NetworkInterfaces']:
            nic = {}
            get_keys(eni, nic, ['Association', 'Groups', 'PrivateIpAddresses', 'SubnetId', 'Ipv6Addresses'])
            instance['network_interfaces'][eni['NetworkInterfaceId']] = nic

        return id, instance

    @staticmethod
    def _identify_user_data_secrets(user_data):
        """
        Parses EC2 user data in order to identify secrets and credentials..
        """
        secrets = {}

        if user_data:
            aws_access_key_regex = re.compile('AKIA[0-9A-Z]{16}')
            aws_secret_access_key_regex = re.compile('[0-9a-zA-Z/+]{40}')
            rsa_private_key_regex = re.compile('(-----BEGIN RSA PRIVATE KEY-----(?s).+?-----END .+?-----)')
            keywords = ['password', 'secret', 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token']

            aws_access_key_list = aws_access_key_regex.findall(user_data)
            if aws_access_key_list:
                secrets['AWS Access Key IDs'] = aws_access_key_list
            aws_secret_access_key_list = aws_secret_access_key_regex.findall(user_data)
            if aws_secret_access_key_list:
                secrets['AWS Secret Access Keys'] = aws_secret_access_key_list
            rsa_private_key_list = rsa_private_key_regex.findall(user_data)
            if rsa_private_key_list:
                secrets['Private Keys'] = rsa_private_key_list
            word_list = []
            for word in keywords:
                if word in user_data.lower():
                    word_list.append(word)
            if word_list:
                secrets['Flagged Words'] = word_list

        return secrets



import asyncio

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.aws.utils import ec2_classic
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently


class ELBv2Facade(AWSBaseFacade):
    regional_load_balancers_cache_locks = {}
    load_balancers_cache = {}

    async def get_load_balancers(self, region: str, vpc: str):
        try:
            await self.cache_load_balancers(region)
            return [load_balancer for load_balancer in self.load_balancers_cache[region] if load_balancer['VpcId'] == vpc]
        except Exception as e:
            print_exception('Failed to get ELBv2 load balancers: {}'.format(e))
            return []

    async def cache_load_balancers(self, region):
        async with self.regional_load_balancers_cache_locks.setdefault(region, asyncio.Lock()):
            if region in self.load_balancers_cache:
                return

            self.load_balancers_cache[region] = \
                await AWSFacadeUtils.get_all_pages('elbv2', region, self.session,
                                                   'describe_load_balancers', 'LoadBalancers')

            for load_balancer in self.load_balancers_cache[region]:
                load_balancer['VpcId'] = \
                    load_balancer['VpcId'] if 'VpcId' in load_balancer and load_balancer['VpcId'] else ec2_classic

            await get_and_set_concurrently(
                [self._get_and_set_load_balancer_attributes], self.load_balancers_cache[region], region=region)

    async def _get_and_set_load_balancer_attributes(self, load_balancer: dict, region: str):
        elbv2_client = AWSFacadeUtils.get_client('elbv2', self.session, region)
        try:
            load_balancer['attributes'] = await run_concurrently(
                lambda: elbv2_client.describe_load_balancer_attributes(
                    LoadBalancerArn=load_balancer['LoadBalancerArn'])['Attributes']
            )
        except Exception as e:
            print_exception('Failed to describe ELBv2 attributes: {}'.format(e))

    async def get_listeners(self, region: str, load_balancer_arn: str):
        return await AWSFacadeUtils.get_all_pages(
            'elbv2', region, self.session, 'describe_listeners', 'Listeners', LoadBalancerArn=load_balancer_arn)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.config.recorders import Recorders
from ScoutSuite.providers.aws.resources.config.rules import Rules
from ScoutSuite.providers.aws.resources.regions import Regions


class Config(Regions):
    _children = [
        (Recorders, 'recorders'),
        (Rules, 'rules')
    ]

    def __init__(self, facade: AWSFacade):
        super(Config, self).__init__('config', facade)



from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import map_concurrently
from ScoutSuite.providers.utils import run_concurrently


class SESFacade(AWSBaseFacade):
    async def get_identities(self, region: str):
        try:
            identity_names = await AWSFacadeUtils.get_all_pages(
                'ses', region, self.session, 'list_identities', 'Identities')

            return await map_concurrently(self._get_identity_dkim_attributes, identity_names, region=region)
        except Exception as e:
            print_exception('Failed to get SES identities: {}'.format(e))
            return []

    async def _get_identity_dkim_attributes(self, identity_name: str, region: str):
        ses_client = AWSFacadeUtils.get_client('ses', self.session, region)
        try:
            dkim_attributes = await run_concurrently(
                lambda: ses_client.get_identity_dkim_attributes(Identities=[identity_name])['DkimAttributes'][
                    identity_name]
            )
        except Exception as e:
            print_exception('Failed to get SES DKIM attributes: {}'.format(e))
            raise
        return identity_name, dkim_attributes

    async def get_identity_policies(self, region: str, identity_name: str):
        ses_client = AWSFacadeUtils.get_client('ses', self.session, region)
        try:
            policy_names = await run_concurrently(
                lambda: ses_client.list_identity_policies(Identity=identity_name)['PolicyNames']
            )
        except Exception as e:
            print_exception('Failed to list SES policies: {}'.format(e))
            policy_names = []

        if len(policy_names) == 0:
            return {}

        try:
            return await run_concurrently(
                lambda: ses_client.get_identity_policies(Identity=identity_name, PolicyNames=policy_names)['Policies']
            )
        except Exception as e:
            print_exception('Failed to get SES policies: {}'.format(e))
            return None



from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class SecurityGroups(AzureResources):
    async def fetch_all(self):
        for raw_group in await self.facade.network.get_network_security_groups():
            id, network_security_group = self._parse_network_security_group(raw_group)
            self[id] = network_security_group

    def _parse_network_security_group(self, network_security_group):
        network_security_group_dict = {}
        network_security_group_dict['id'] = get_non_provider_id(network_security_group.id)
        network_security_group_dict['name'] = network_security_group.name
        network_security_group_dict['location'] = network_security_group.location
        network_security_group_dict['provisioning_state'] = network_security_group.provisioning_state
        network_security_group_dict['resource_guid'] = network_security_group.resource_guid
        network_security_group_dict['type'] = network_security_group.type
        network_security_group_dict['etag'] = network_security_group.etag
        network_security_group_dict['tags'] = network_security_group.tags
        network_security_group_dict['additional_properties'] = network_security_group.additional_properties
        network_security_group_dict['network_interfaces'] = network_security_group.network_interfaces

        network_security_group_dict['security_rules'] = self._parse_security_rules(network_security_group)

        network_security_group_dict['subnets'] = {}
        if network_security_group.subnets:
            for subnet in network_security_group.subnets:
                identifier = get_non_provider_id(subnet.id)
                network_security_group_dict['subnets'][identifier] = {'id': identifier}

        # FIXME this is broken and badly implemented (not efficient at all)
        # exposed_ports = self._parse_exposed_ports(network_security_group)
        # network_security_group_dict['exposed_ports'] = exposed_ports
        # network_security_group_dict['exposed_port_ranges'] = self._format_ports(exposed_ports)

        return network_security_group_dict['id'], network_security_group_dict

    def _parse_security_rules(self, network_security_group):
        security_rules = {}
        # custom rules
        for sr in network_security_group.security_rules:
            security_rule_id, security_rule_dict = self._parse_security_rule(sr)
            security_rules[security_rule_id] = security_rule_dict
        # default rules
        for sr in network_security_group.default_security_rules:
            security_rule_id, security_rule_dict = self._parse_security_rule(sr, default=True)
            security_rules[security_rule_id] = security_rule_dict
        return security_rules

    def _parse_security_rule(self, rule, default=False):
        security_rule_dict = {}
        security_rule_dict['id'] = rule.id
        security_rule_dict['name'] = rule.name
        security_rule_dict['allow'] = rule.access == "Allow"
        security_rule_dict['priority'] = rule.priority
        security_rule_dict['description'] = rule.description
        security_rule_dict['provisioning_state'] = rule.provisioning_state

        security_rule_dict['protocol'] = rule.protocol
        security_rule_dict['direction'] = rule.direction

        source_address_prefixes = self._merge_prefixes_or_ports(rule.source_address_prefix,
                                                                rule.source_address_prefixes)
        security_rule_dict['source_address_prefixes'] = source_address_prefixes

        source_port_ranges = self._merge_prefixes_or_ports(rule.source_port_range, rule.source_port_ranges)
        security_rule_dict['source_port_ranges'] = source_port_ranges
        security_rule_dict['source_ports'] = self._parse_ports(source_port_ranges)

        destination_address_prefixes = self._merge_prefixes_or_ports(rule.destination_address_prefix,
                                                                     rule.destination_address_prefixes)
        security_rule_dict['destination_address_prefixes'] = destination_address_prefixes

        destination_port_ranges = self._merge_prefixes_or_ports(rule.destination_port_range,
                                                                rule.destination_port_ranges)
        security_rule_dict['destination_port_ranges'] = destination_port_ranges
        security_rule_dict['destination_ports'] = self._parse_ports(destination_port_ranges)

        security_rule_dict['etag'] = rule.etag

        security_rule_dict['default'] = default

        return security_rule_dict['id'], security_rule_dict


    def _parse_ports(self, port_ranges):
        # FIXME this is inefficient
        ports = set()
        for pr in port_ranges:
            if pr == "*":
                for p in range(0, 65535 + 1):
                    ports.add(p)
                break
            elif "-" in pr:
                lower, upper = pr.split("-")
                for p in range(int(lower), int(upper) + 1):
                    ports.add(p)
            else:
                ports.add(int(pr))
        ports = list(ports)
        ports.sort()
        return ports

    def _parse_exposed_ports(self, network_security_group):
        exposed_ports = set()

        # Sort by priority.
        rules = network_security_group.default_security_rules + network_security_group.security_rules
        rules.sort(key=lambda x: x.priority, reverse=True)

        for sr in rules:
            if sr.direction == "Inbound" and (sr.source_address_prefix == "*"
                                              or sr.source_address_prefix == "Internet"):
                port_ranges = self._merge_prefixes_or_ports(sr.destination_port_range,
                                                                     sr.destination_port_ranges)
                ports = self._parse_ports(port_ranges)
                if sr.access == "Allow":
                    for p in ports:
                        exposed_ports.add(p)
                else:
                    for p in ports:
                        exposed_ports.discard(p)
        exposed_ports = list(exposed_ports)
        exposed_ports.sort()
        return exposed_ports

    def _merge_prefixes_or_ports(self, port_range, port_ranges):
        port_ranges = port_ranges if port_ranges else []
        if port_range:
            port_ranges.append(port_range)
        return port_ranges

    def _format_ports(self, ports):
        # FIXME this is inefficient
        port_ranges = []
        start = None
        for i in range(0, 65535 + 1):
            if i in ports:
                if not start:
                    start = i
            else:
                if start:
                    if i - 1 == start:
                        port_ranges.append(str(start))
                    else:
                        port_ranges.append(str(start) + "-" + str(i - 1))
                    start = None
        return port_ranges




from azure.mgmt.keyvault import KeyVaultManagementClient
from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception


class KeyVaultFacade:
    def __init__(self, credentials, subscription_id):
        self._client = KeyVaultManagementClient(credentials, subscription_id)

    async def get_key_vaults(self):
        try:
            return await run_concurrently(
                lambda: list(self._client.vaults.list_by_subscription()))
        except Exception as e:
            print_exception('Failed to retrieve key vaults: {}'.format(e))
            return []



from ScoutSuite.providers.utils import run_concurrently

class GCPFacadeUtils:
    @staticmethod
    def _get_all(resources, resource_key: str, request, resources_group):
        while request is not None:
            response = request.execute()
            resources.extend(response.get(resource_key, []))
            request = resources_group.list_next(previous_request=request, previous_response=response)

    @staticmethod
    async def get_all(resource_key: str, request, resources_group):
        resources = []
        await run_concurrently(
            lambda: GCPFacadeUtils._get_all(resources, resource_key, request, resources_group)
        )
        return resources
